;;; Copyright (C) 2002 Gerd Moellmann <gerd.moellmann@t-online.de>
;;; All rights reserved.
;;;
;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions
;;; are met:
;;;
;;; 1. Redistributions of source code must retain the above copyright
;;;    notice, this list of conditions and the following disclaimer.
;;; 2. Redistributions in binary form must reproduce the above copyright
;;;    notice, this list of conditions and the following disclaimer in the
;;;    documentation and/or other materials provided with the distribution.
;;; 3. The name of the author may not be used to endorse or promote
;;;    products derived from this software without specific prior written
;;;    permission.
;;;
;;; THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
;;; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
;;; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
;;; LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
;;; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
;;; OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
;;; BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
;;; LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
;;; USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
;;; DAMAGE.

;;; To Do
;;;
;;; GF is actually non-accessor GF.  Clean this up.
;;; (setf symbol-value) should be handled like (setf fdefinition)

(ext:file-comment "$Header: /u1/gerd/cmucl/src/mypcl/RCS/info.lisp,v 1.31 2002/12/02 12:29:26 gerd Exp $")

(in-package "PCL")

(defstruct class-info
  ;;
  ;; List of SLOT-INFO structures for each slot.
  (slots () :type list)
  ;;
  ;; The class' metaclass.
  (metaclass nil :type symbol)
  ;;
  ;; List of slot reader and writer function names.
  (accessors () :type list))


(defstruct slot-info
  (name nil :type symbol)
  ;;
  ;; Specified slot allocation.or :INSTANCE.
  (allocation :instance :type (member :class :instance))
  ;;
  ;; Specified slot type or T.
  (type t :type (or symbol list)))


(defstruct gf-info
  ;;
  ;; Number of required parameters.
  (nreq 0 :type fixnum)
  ;;
  ;; True if this generic function has key, optional, or rest parameters.
  (applyp nil :type boolean)
  ;;
  ;; READER if this generic function has a slot reader method.
  ;; WRITER if this gf has a slot writer method.
  ;; NIL if this is an ordinary generic function.
  (type nil :type (member nil reader writer))
  ;;
  ;; List of class names for slot accessor.  A class C occurring in
  ;; this list means that this is a slot accessor for some slot of C.
  (classes () :type list))


;;;
;;; Declare some new INFO type, under class PCL, plus nicer accessors.
;;;
(ext:define-info-class pcl)
(ext:define-info-type pcl class (or null class-info) nil)
(ext:define-info-type pcl gf (or null gf-info) nil)

(declaim (inline class-info (setf class-info)
		 gf-info (setf gf-info)))

(defun class-info (class-name)
  (ext:info pcl class class-name))

(defun (setf class-info) (new-value class-name)
  (setf (ext:info pcl class class-name) new-value))

(defun gf-info (gf-name)
  (ext:info pcl gf gf-name))

(defun (setf gf-info) (new-value gf-name)
  (setf (ext:info pcl gf gf-name) new-value))


;;; ******************
;;; Classes   ********
;;; ******************

;;;
;;; Set compile-time info for class CLASS-NAME.  Called from
;;; EXPAND-DEFCLASS.
;;;
;;; METACLASS is the metaclass of the declared class.  SLOTS is a
;;; list of slot specs as they appear in DEFCLASS.
;;;
(defun set-class-info (class-name metaclass slots)
  (let ((readers ())
	(writers ())
	(slot-infos ())
	(class-info (class-info class-name)))
    (when class-info
      (loop for a in (class-info-accessors class-info)
	    as info = (gf-info a) do
	      (setf (gf-info-classes info)
		    (delete class-name (gf-info-classes info)))))
    ;;
    (dolist (slot slots)
      (typecase slot
	(list
	 (loop for (key value) on (cdr slot) by #'cddr do
		 (case key
		   (:reader
		    (push value readers))
		   (:writer
		    (push value writers))
		   (:accessor
		    (push value readers)
		    (push `(setf ,value) writers))))
	 (push (make-slot-info
		:name (car slot)
		:allocation (getf (cdr slot) :allocation :instance)
		:type (getf (cdr slot) :type t))
	       slot-infos))
	(t
	 (push (make-slot-info :name slot) slot-infos))))
    ;;
    (setf (class-info class-name)
	  (make-class-info :slots (nreverse slot-infos)
			   :accessors (append readers writers)
			   :metaclass metaclass))
    ;;
    (flet ((add-class (accessors type)
	     (loop for a in accessors as info = (gf-info a)
		   if (null info) do
		     (setf (gf-info a)
			   (make-gf-info :nreq (ecase type
						 (reader 1)
						 (writer 2))
					 :applyp nil
					 :classes (list class-name)
					 :type type))
		   else do
		     (setf (gf-info-type info) type)
		     (push class-name (gf-info-classes info)))))
      (add-class readers 'reader)
      (add-class writers 'writer))))

;;;
;;; Decide what type slot SLOT-NAME in CLASS-OR-NAME has.
;;; CLASS-OR-NAME can be a class or class name.
;;;
(defun decide-slot-type (class-or-name slot-name)
  (or (let* ((class-info (class-info (if (symbolp class-or-name)
					 class-or-name
					 (class-name class-or-name))))
	     (slot-info (when class-info
			  (find slot-name (class-info-slots class-info)
				:key #'slot-info-name :test #'eq))))
	(when slot-info
	  (slot-info-type slot-info)))
      (let* ((class (if (symbolp class-or-name)
			(find-class class-or-name nil)
			class-or-name))
	     (slotd (when class
		      (find-slot-definition class slot-name))))
	(when slotd
	  (slot-definition-type slotd)))))

;;;
;;; Return T if CLASS-NAME names a STD-CLASS, that is, either a
;;; STANDARD-CLASS or FUNCALLABLE-STANDARD-CLASS in normal MOP spaak.
;;;
;;; Structure classes are defined at compile-time, see the EVAL in
;;; EXPAND-DEFCLASS.  If it's a structure, the class must be defined,
;;; and be a structure class.
;;;
(defun info-std-class-p (class-name)
  (let ((class (find-class class-name nil)))
    (if class
	(std-class-p class)
	(class-info class-name))))


;;; **************
;;; Slots  *******
;;; **************

;;;
;;; Return true if NAME should be considerd a slot reader/writer name.
;;;
(defun info-accessor-p (name)
  (or (let ((info (gf-info name)))
	(when info
	  (not (null (gf-info-type info)))))
      (when (and (eq *boot-state* 'complete)
		 (fboundp name))
	(let ((gf (fdefinition name))
	      (accessorp nil))
	  (when (generic-function-p gf)
	    (loop for method in (generic-function-methods gf)
		  when (standard-accessor-method-p method) do
		    (setq accessorp t)
		    (let ((type (if (standard-reader-method-p method)
				    'reader 'writer)))
		      (setf (gf-info name)
			    (make-gf-info :nreq (ecase type
						  (reader 1)
						  (writer 2))
					  :applyp nil
					  :type type
					  :classes ())))
		    (loop-finish)))
	  accessorp))))

;;;
;;; Decide if a call to GF-NAME should be optimized as a slot
;;; reader/writer call of type TYPE.  TYPE must be one of the symbols
;;; READER or WRITER.
;;;
(defun decide-optimize-accessor-p (gf-name type)
  (let ((info (gf-info gf-name)))
    (when info
      (eq type (gf-info-type info)))))

;;;
;;; Decide if access to slot SLOT-NAME in class CLASS-NAME should be
;;; optimized.  TYPE is the type of access, one of READER/WRITER/ALL.
;;;
;;; This function is called from COMPUTE-PV-SLOT and from the code
;;; actually generating the code for optimized slot access.  It's
;;; suboptimal but okay if this function errs: either we'll be doing a
;;; superfluous optimization or prevent a possible optimization.  The
;;; PV code must anyway be able to cope with optimizations that must
;;; be "undone" by setting pv-values to nil.
;;;
;;; It's worthwhile optimizing slot access if we know or suspect at
;;; method compilation time that slots are accessed in the standard
;;; way, which means that there is no non-standard
;;; SLOT-VALUE-USING-CLASS method defined.
;;; 
;;; If we know that there is or will be such a method, it's not worth
;;; optimizing.
;;;
(defun decide-optimize-slot-p (class-name slot-name type)
  (or (let ((class-info (class-info class-name)))
	(and class-info
	     (member slot-name (class-info-slots class-info)
		     :key #'slot-info-name :test #'eq)
	     ;; FUNCALLABLE-STANDARD-CLASS might be ok, too, I haven't
	     ;; checked.  It's not terribly important, though.
	     (eq (class-info-metaclass class-info) 'standard-class)))
      (let* ((class (find-class class-name nil))
	     (slotd (and class (find-slot-definition class slot-name))))
	(when slotd
	  (slot-accessor-std-p slotd type)))))

;;;
;;; Decide if SLOT-NAME should be considered a class slot in class
;;; CLASS-NAME.
;;;
(defun decide-class-slot-p (class-name slot-name)
  (or (let* ((class-info (class-info class-name))
	     (slot-info (and class-info
			     (find slot-name (class-info-slots class-info)
				   :key #'slot-info-name :test #'eq))))
	(when slot-info
	  (eq :class (slot-info-allocation slot-info))))
      (let* ((class (find-class class-name nil))
	     (slotd (and class (find-slot-definition class slot-name))))
	(when slotd
	  (eq :class (slot-definition-allocation slotd))))))


;;; ***************************
;;; Generic Functions   *******
;;; ***************************

;;;
;;; Record compile-time information about the declaration of
;;; generic function GF-NAME with lambda list LAMBDA-LIST.
;;;
(defun set-gf-info (gf-name lambda-list)
  (multiple-value-bind (required optional restp rest keyp keys
				 allow-other-keys-p)
      (parse-generic-function-lambda-list lambda-list)
    (declare (ignore rest keys))
    (setf (gf-info gf-name)
	  (make-gf-info :nreq (length required)
			;; Like in ARG-INFO-APPLYP.
			:applyp (or restp keyp (not (null optional))
				    allow-other-keys-p)))))

;;;
;;; Return true if NAME is the name of a known generic function.
;;;
(defun info-gf-name-p (name)
  (when (or (symbolp name)
	    (and (consp name)
		 (eq (car name) 'setf)
		 (consp (cdr name))
		 (symbolp (cadr name))
		 (null (cddr name))))
    (or (gf-info name)
	(when (and (fboundp name)
		   (eq *boot-state* 'complete))
	  (let ((gf (gdefinition name)))
	    (when (standard-generic-function-p gf)
	      (set-gf-info name (generic-function-lambda-list gf))
	      t))))))

;;;
;;; Called to update the info database and pv caches when function
;;; NAME gets redefined to NEWDEF.  If NAME is a generic function,
;;; flush pv caches referring to it.
;;;
(defun %check-gf-redefinition (name newdef)
  (when (eq *boot-state* 'complete)
    (let ((gf (and (fboundp name) (fdefinition name))))
      (when (and gf
		 (generic-function-p gf)
		 (not (generic-function-p newdef)))
	(setf (gf-info name) nil)
	(update-accessor-pvs 'removed-gf gf)
	(update-pv-calls-for-gf gf 'removed-gf)))))

(push #'%check-gf-redefinition lisp::*setf-fdefinition-hook*)

;;; end of info.lisp
